/*
	$Id: Collection.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Collection.hpp>
#include <librfc/CollectionException.hpp>


USE_RFC_NAMESPACE

Collection::Collection( long lDeleteMembersOnDestuction )
			: _lDeleteMembersOnDestuction( lDeleteMembersOnDestuction )
{
	_pFirst = NULL;
	_pLast = NULL;
	_ulMembers = 0;
#if defined (_WINDOWS)
	InitializeCriticalSection( &_cs );
#elif defined (LINUX)
	_pmx = new pthread_mutex_t;
	if( pthread_mutex_init( _pmx, NULL ) )
		Throw( "Collection", 1 );
#endif
}

Collection::~Collection()
{
	removeAll( _lDeleteMembersOnDestuction );

#if defined (_WINDOWS)
	DeleteCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_destroy( _pmx );
	delete _pmx;
#endif
}

void Collection::addTail( Collectable* pitem )
{
#if defined (_WINDOWS)
	EnterCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_lock( _pmx );
#endif

	try
	{
		pitem->addToCollectionTail( this );
		_ulMembers++;
	}
	catch( CollectionException& /*e*/ )
	{
#if defined (_WINDOWS)
		LeaveCriticalSection( &_cs );
#elif defined (LINUX)
		pthread_mutex_unlock( _pmx );
#endif
		throw;
	}

#if defined (_WINDOWS)
	LeaveCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_unlock( _pmx );
#endif
}

void Collection::addHead( Collectable* pitem )
{
#if defined (_WINDOWS)
	EnterCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_lock( _pmx );
#endif

	try
	{
		pitem->addToCollectionHead( this );
		_ulMembers++;
	}
	catch( CollectionException& /*e*/ )
	{
#if defined (_WINDOWS)
		LeaveCriticalSection( &_cs );
#elif defined (LINUX)
		pthread_mutex_unlock( _pmx );
#endif
		throw;
	}

#if defined (_WINDOWS)
	LeaveCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_unlock( _pmx );
#endif
}

Collectable* Collection::next( Collectable* pCurrent )
{
	if( pCurrent == NULL )
	{
		return _pFirst;
	}
	else
	{
		return pCurrent->next( this );
	}
}

Collectable* Collection::previous( Collectable* pCurrent )
{
	if( pCurrent == NULL )
	{
		return _pLast;
	}
	else
	{
		return pCurrent->previous( this );
	}
}

Collectable* Collection::find( const String& strKey, Collectable* pCurrent )
{
	Collectable* pFound = NULL;

	if( pCurrent == NULL )
	{
		pCurrent = _pFirst;
	}
	else
	{
		pCurrent = pCurrent->next( this );
	}

	while( ( pCurrent != NULL ) && ( pFound == NULL ) )
	{
		if( pCurrent->_strKey == strKey )
		{
			pFound = pCurrent;
		}
		else
		{
			pCurrent = pCurrent->next( this );
		}
	}

	return pFound;
}

void Collection::remove ( Collectable* pRemove, long bDelete )
{
/*#if defined (_WINDOWS)
	EnterCriticalSection( &_cs );
#endif

	if( pRemove->_pNext == NULL )
	{
		_pLast = pRemove->_pPrevious;
	}
	if( pRemove->_pPrevious != NULL )
	{
		pRemove->_pPrevious->_pNext = pRemove->_pNext;
	}

	if( pRemove->_pPrevious == NULL )
	{
		_pFirst = pRemove->_pNext;
	}
	if( pRemove->_pNext != NULL )
	{
		pRemove->_pNext->_pPrevious = pRemove->_pPrevious;
	}

	if( bDelete )
	{
		delete pRemove;
	}
	else
	{
		pRemove->_pNext = pRemove->_pPrevious = pRemove->_pParent = NULL;
	}

	_ulMembers--;

#if defined (_WINDOWS)
	LeaveCriticalSection( &_cs );
#endif*/

#if defined (_WINDOWS)
	EnterCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_lock( _pmx );
#endif

	try
	{
		pRemove->removeFrom( this );
		_ulMembers--;

		// The question 'Should I delete you ?' is for the collection, based
		//  upon the bDelete Argument.
		// The question 'Can I delete you ?' is for the collectable.
		if( bDelete )
		{
			if( pRemove->canIDeleteYou() )
			{
				delete pRemove;
			}
		}

	}
	catch( CollectionException& /*e*/ )
	{
#if defined (_WINDOWS)
		LeaveCriticalSection( &_cs );
#elif defined (LINUX)
		pthread_mutex_unlock( _pmx );
#endif
		throw;
	}

#if defined (_WINDOWS)
	LeaveCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_unlock( _pmx );
#endif
}

void Collection::removeAll( long bDelete )
{
	Collectable* pCurrent = _pFirst;
	Collectable* pNext;

	while( pCurrent != NULL )
	{
		pNext = pCurrent->next( this );
		remove( pCurrent, bDelete );
		pCurrent = pNext;
	}
}

unsigned long Collection::count() const
{
	return _ulMembers;
}
