/*
	$Id: DBPoolMgr.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/DBPoolMgr.hpp>
#include <librfc/DBPoolException.hpp>
#include <librfc/DatabaseException.hpp>

#include <librfc/Socket.hpp>


USE_RFC_NAMESPACE

DBPoolMgr::DBPoolMgr()
	: _pdbFirst( NULL ),
	_pdbLast( NULL ),
	_lInitialSize( 0 ),
	_lMaxSize( -1 ),
	_lCurrentSize( 0 ),
	_bResizeable( false )
{
#if defined (_DEBUG)
	_lTotal = 0;
	_lInUse = 0;
#endif
	_ulFailureCount = 0;
#if defined (_WINDOWS)
	InitializeCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_init( &_mx, NULL );
#endif
}

DBPoolMgr::~DBPoolMgr()
{
	if( _pdbFirst != NULL )
		delete _pdbFirst;

#if defined (_WINDOWS)
	DeleteCriticalSection( &_cs );
#elif defined (LINUX)
#endif
}

long DBPoolMgr::addPooledDB( PooledDB* pdb )
{
	if( pdb == NULL )
		return -1;

	long lRtn = 0;

	if( _pdbFirst == NULL )
	{
		_pdbFirst = pdb;
		_pdbLast = pdb;
	}
	else
	{
		_pdbLast->_pdbNext = pdb;
		_pdbLast = pdb;
	}

	_lCurrentSize++;

#ifdef _DEBUG
 _lTotal++;
#endif

	return lRtn;
}

long DBPoolMgr::setResizeable( const String& strSource, const String& strUserId, const String& strPassword,
						long lInitialSize, long lMaxSize )
{
	long lRtn = 0;

	_strDSN = strSource;
	_strUserId = strUserId;
	_strPassword = strPassword;
	_lInitialSize = lInitialSize;
	_lMaxSize = lMaxSize;
	_bResizeable = true;

	try
	{
		for( int nI = 0; nI < _lInitialSize; nI++ )
		{
			PooledDB* pdb = new PooledDB( _strDSN, _strUserId, _strPassword );
			addPooledDB( pdb );
		}
	}
	catch( const DBPoolException& e )
	{
		lRtn = -1;
	}

	return lRtn;
}

long DBPoolMgr::addPooledDB( const char* pszSource, const char* pszUserId, const char* pszPassword )
{
	long lRtn = 0;

	try
	{
		PooledDB* pdb = new PooledDB( pszSource, pszUserId, pszPassword );
		addPooledDB( pdb );
	}
	catch( const DBPoolException& e )
	{
		lRtn = -1;
	}

	return lRtn;
}

PooledDB* DBPoolMgr::getPooledDB()
{
#if defined (_WINDOWS)
	EnterCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_lock( &_mx );
#endif

	PooledDB* pdbFound = NULL;
	PooledDB* pdbCurrent = _pdbFirst;

	while( ( pdbFound == NULL ) && ( pdbCurrent != NULL ) )
	{
		if( !pdbCurrent->_lInUse )
		{
			pdbFound = pdbCurrent;
			pdbFound->_lInUse = 1;
		}
		else
		{
			pdbCurrent = pdbCurrent->_pdbNext;
		}
	}

	if( ( pdbFound == NULL ) &&
		( _bResizeable ) &&
		( _lCurrentSize < _lMaxSize ) )
	{
		try
		{
			pdbFound = new PooledDB( _strDSN, _strUserId, _strPassword );
			addPooledDB( pdbFound );
			pdbFound->_lInUse = 1;
		}
		catch( const DBPoolException& e )
		{
		}
		catch( const DatabaseException& e )
		{
		}
		catch( ... )
		{
		}
	}

#if defined (_WINDOWS)
	LeaveCriticalSection( &_cs );
#elif defined (LINUX)
	pthread_mutex_unlock( &_mx );
#endif

	if( pdbFound->refresh() )
		pdbFound->refresh( _strDSN, _strUserId, _strPassword );

	return pdbFound;
}

void DBPoolMgr::freePooledDB( PooledDB* pdbFree )
{
	pdbFree->_lInUse = 0;
}

void DBPoolMgr::refresh()
{
	if( !_bResizeable )
		return;

	PooledDB* pdbCurrent = _pdbFirst;

	while( pdbCurrent != NULL )
	{
		pdbCurrent->setRefresh();
	}
}

void DBPoolMgr::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl
		<< "Src=" << ( _strDSN.isNull() ? "(null)" : (const char*)_strDSN )
		<< " Uid=" << ( _strUserId.isNull() ? "(null)" : (const char*)_strUserId )
		<< " Pwd=" << ( _strPassword.isNull() ? "(null)" : (const char*)_strPassword )
		<< std::endl;
}

void DBPoolMgr::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n"
		<< "Src=" << ( _strDSN.isNull() ? "(null)" : (const char*)_strDSN )
		<< " Uid=" << ( _strUserId.isNull() ? "(null)" : (const char*)_strUserId )
		<< " Pwd=" << ( _strPassword.isNull() ? "(null)" : (const char*)_strPassword )
		<< "\n";
}
