/*
	$Id: Database.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Database.hpp>
#include <librfc/RFCAssert.h>
#include <librfc/Socket.hpp>

#include <librfc/rfcversion.h>

#include <string.h>
#if defined (LINUX)
#	include <unistd.h>
#endif


USE_RFC_NAMESPACE

Database::Database()
{
	_bConnValid = 0;
	_bIsOpen = 0;
	_bShouldBeAutocommit = 0;
	_bInTransaction = 0;

	_szSource[0] = 0;
	_szUser[0] = 0;
	_szAuth[0] = 0;

	_hConn = SQL_NULL_HANDLE;
}

Database::~Database()
{
	if( _bInTransaction )
		rollback();

	if( _bIsOpen )
		close();

	if( _bConnValid )
	{
		SQLFreeHandle( SQL_HANDLE_DBC, _hConn );
	}
}

SQLHANDLE Database::handle()
{
	return _hConn;
}

long Database::open( const char* pszSource, const char* pszUser, const char* pszAuth )
{
	long lRtn = 0;
	SQLRETURN rc;

	if( _ODBCEnvironment.isEnvValid() )
	{
		strcpy( _szSource, pszSource );
		strcpy( _szUser, pszUser );
		strcpy( _szAuth, pszAuth );

		_bConnValid = TRUE;

		rc = SQLAllocHandle( SQL_HANDLE_DBC, _ODBCEnvironment.handle(), &_hConn );
		if( ( rc == SQL_SUCCESS ) ||
			( rc == SQL_SUCCESS_WITH_INFO ) )
		{
			rc = SQLConnect( _hConn,
										(unsigned char*)_szSource,
										strlen( _szSource ),
										(unsigned char *)_szUser,
										strlen( _szUser ),
										(unsigned char *)_szAuth,
										strlen( _szAuth ) );
			if( ( rc != SQL_SUCCESS ) &&
				( rc != SQL_SUCCESS_WITH_INFO ) )
			{
				LONG lOutInt;
				SHORT nTextOut;
				SQLTCHAR szOutState[6];
				SQLTCHAR szOutMessage[400];

				ASSERT( 0 );

				SQLGetDiagRec( SQL_HANDLE_DBC, _hConn, 1, szOutState,
								(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

				_strLastError = (const char*)szOutMessage;

				_bConnValid = FALSE;
				lRtn = -1;
			}
			else
			{
				_bIsOpen = 1;
			}
		}
		else
		{
			LONG lOutInt;
			SHORT nTextOut;
			SQLTCHAR szOutState[6];
			SQLTCHAR szOutMessage[400];

			ASSERT( 0 );

			SQLGetDiagRec( SQL_HANDLE_DBC, _hConn, 1, szOutState,
							(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

			_strLastError = (const char*)szOutMessage;

			_bConnValid = FALSE;
			lRtn = -2;
		}

	}
	else
		lRtn = -3;

	return lRtn;
}

long Database::close()
{
	long lRtn = 0;

	if( _bIsOpen )
	{
		SQLDisconnect( _hConn );
		_bIsOpen = FALSE;
	}

	if( _bConnValid )
	{
		SQLFreeHandle( SQL_HANDLE_DBC, _hConn );
		_bConnValid = FALSE;
	}

	_szSource[0] = 0;
	_szUser[0] = 0;
	_szAuth[0] = 0;

	return lRtn;
}

long Database::execute( const char* pszStatement )
{
	long lRtn = 0;
	SQLRETURN rc;
	SQLHANDLE hStmt;

	rc = SQLAllocHandle( SQL_HANDLE_STMT, _hConn, &hStmt );
	if( ( rc != SQL_SUCCESS )&&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_STMT, hStmt, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = -4;
		return lRtn;		// Exit Point
	}

	rc = SQLExecDirect( hStmt, (unsigned char*)pszStatement, strlen( pszStatement ) );
	if( ( rc != SQL_SUCCESS ) &&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_STMT, hStmt, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = -1;
	}

	SQLLEN nRowCount;
	SQLRowCount( hStmt, &nRowCount );
	_lRowCount = nRowCount;

	rc = SQLFreeStmt( hStmt, SQL_DROP );
	if( ( rc != SQL_SUCCESS ) &&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_STMT, hStmt, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = 1;
	}
//	SQLFreeHandle( SQL_HANDLE_STMT, hStmt );

	return lRtn;
}

long Database::enableAutoCommit( long lEnable )
{
	long lRtn = 0;
	SQLRETURN rc;

	if( lEnable )
	{
		rc = SQLSetConnectOption( _hConn, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON );
		_bShouldBeAutocommit = 1;
	}
	else
	{
		rc = SQLSetConnectOption( _hConn, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF );
		_bShouldBeAutocommit = 0;
	}
	if( ( rc != SQL_SUCCESS ) &&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_DBC, _hConn, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = -1;
	}

	return lRtn;
}

void Database::begin()
{
	if( _bShouldBeAutocommit )
		SQLSetConnectOption( _hConn, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF );
	_bInTransaction = 1;
}

long Database::commit()
{
	long lRtn = 0;
	SQLRETURN rc;

	rc = SQLTransact( _ODBCEnvironment.handle(), _hConn, SQL_COMMIT );
	if( ( rc != SQL_SUCCESS ) &&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_DBC, _hConn, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = -1;
	}

	if( _bShouldBeAutocommit )
		SQLSetConnectOption( _hConn, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON );
	_bInTransaction = 0;

	return lRtn;
}

long Database::rollback()
{
	long lRtn = 0;
	SQLRETURN rc;

	rc = SQLTransact( _ODBCEnvironment.handle(), _hConn, SQL_ROLLBACK );
	if( ( rc != SQL_SUCCESS ) &&
		( rc != SQL_SUCCESS_WITH_INFO ) )
	{
		LONG lOutInt;
		SHORT nTextOut;
		SQLTCHAR szOutState[6];
		SQLTCHAR szOutMessage[400];

		ASSERT( 0 );

		SQLGetDiagRec( SQL_HANDLE_DBC, _hConn, 1, szOutState,
						(SQLINTEGER*)&lOutInt, szOutMessage, 400, &nTextOut );

		_strLastError = (const char*)szOutMessage;

		lRtn = -1;
	}

	if( _bShouldBeAutocommit )
		SQLSetConnectOption( _hConn, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_ON );
	_bInTransaction = 0;

	return lRtn;
}

long Database::rowCount() const
{
	return _lRowCount;
}

char* Database::DeQuote( char* pszTgt, const char* pszSrc )
{
	int nIndexSrc = 0;
	int nIndexTgt = 0;
	char cCurrent;
	int nSize = strlen( pszSrc ) * 2;

	if( !nSize )
	{
		pszTgt[0] = 0;
		return pszTgt;
	}

	char* pszTemp = new char[nSize];

	memset( pszTemp, 0, nSize * sizeof( char ) );
	
	cCurrent = pszSrc[nIndexSrc];
	while( cCurrent )
	{
		if( cCurrent == '\'' )
			pszTemp[nIndexTgt++] = cCurrent;

		pszTemp[nIndexTgt++] = cCurrent;
		cCurrent = pszSrc[++nIndexSrc];
	}

	strcpy( pszTgt, pszTemp );
	delete pszTemp;


	return pszTgt;
}

void Database::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl << "Src=" << _szSource
		<< " Usr=" << _szUser << "Pwd=" << _szAuth << std::endl;
}

void Database::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n" << "Src=" << _szSource
		<< " Usr=" << _szUser << "Pwd=" << _szAuth << "\n";
}
