/*
	$Id: Date.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Date.hpp>
#include <librfc/DateTimeException.hpp>
#include <librfc/Socket.hpp>

#include <time.h>


USE_RFC_NAMESPACE

const char* pszMonthS[] = { "Jan", "Feb", "Mar", "Apr",
							"May", "Jun", "Jul", "Aug",
							"Sep", "Oct", "Nov", "Dec" };
const long lMonthDayS[] = { 31, -1, 31, 30,
							31, 30, 31, 31,
							30, 31, 30, 31 };
const char* pszDateLegalDelimiters = " -/\\.";

const char* pszLongMonthS[] = { "January", "February", "March", "April",
								"May", "June", "July", "August",
								"September", "October", "November", "December" };

const char* pszDayS[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

const char* pszLongDayS[] = { "Sunday", "Monday", "Tuesday", "Wednesday",
								 "Thursday", "Friday", "Saturday" };

const char* pszMaskYear = "Y";
const char* pszMaskMonth = "M";
const char* pszMaskDay = "D";
const char* pszMaskWeekday = "W";

#define IS_LEAP_YEAR(x) (x % 4) == 0

String Date::shortMonth( long lIndex )
{
	if( ( lIndex < 0 ) || ( lIndex > 11 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}
	return String( pszMonthS[lIndex] );
}

String Date::longMonth( long lIndex )
{
	if( ( lIndex < 0 ) || ( lIndex > 11 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}
	return String( pszLongMonthS[lIndex] );
}

Date::Date()
{
	_strFormat = GetFCCString( SE_DEFAULT_DATE_FORMAT );
	_lYear = 0;
	_lMonth = 0;
	_lDay = 0;
}

Date::Date( long lYear, long lMonth, long lDay )
{
	_strFormat = GetFCCString( SE_DEFAULT_DATE_FORMAT );
	year( lYear );
	month( lMonth );
	day( lDay );
}

Date::Date( const String& strDate, const String& strFormat )
{
	_strFormat = strFormat;

	unsigned long ulMaskLength = _strFormat.length();

	if( strDate.length() != ulMaskLength )
	{
		DateTimeException e( DTE_MASK_MISMATCH );
		throw e;
	}

	for( unsigned long ulIndex = 0; ulIndex < ulMaskLength; ulIndex++ )
	{
		if( strchr( pszDateLegalDelimiters, _strFormat.at( ulIndex ) ) != NULL )
		{
			if( _strFormat.at( ulIndex ) != strDate.at( ulIndex ) )
			{
				DateTimeException e( DTE_MASK_MISMATCH );
				throw e;
			}
		}
	}

	long lYear = (long)strDate.mask( _strFormat.createMask( pszMaskYear ) ).trim();
	if( lYear < 100 )
	{
		long lStartYear = GetFCCLong( FCC_TWODIGITSTARTYEAR );
		long lCutoff = lStartYear % 100;
		long lCentury = ( lStartYear / 100 ) * 100;

		if( lYear < lCutoff )
		{
			lYear += 100;
		}

		lYear += lCentury;
	}
	year( lYear );

	String strMonth = strDate.mask( _strFormat.createMask( pszMaskMonth ) ).trim();
	long lMonth = 0;
	if( strMonth.length() > 2 )
	{
		for( int nMonth = 0; nMonth < 12; nMonth++ )
		{
			if( !strMonth.compareFor( pszMonthS[nMonth], 3, 1 ) )
			{
				lMonth = nMonth + 1;
				break;
			}
		}
	}
	else
	{
		lMonth = (long)strMonth;
	}

	String strDay = strDate.mask( _strFormat.createMask( pszMaskDay ) ).trim();
	if( strDay.length() == 3 )
	{
		struct tm tm;
		memset( &tm, 0, sizeof( tm ) );
		tm.tm_year = _lYear - 1900;
		tm.tm_yday = (long)strDay;
		if( mktime( &tm ) == -1 )
		{
			DateTimeException e( DTE_INVALID_VALUE );
			throw e;
		}
		month( tm.tm_mon + 1 );
		day( tm.tm_mday );

	}
	else
	{
		month( lMonth );
		day( (long)strDay );
	}
}

Date::Date( const Date& date )
{
	_strFormat = date._strFormat;
	_lYear = date._lYear;
	_lMonth = date._lMonth;
	_lDay = date._lDay;
}

Date::~Date()
{
}

const Date& Date::operator= ( const String& strDate )
{
	*this = Date( strDate, _strFormat );

	return *this;
}

const Date& Date::operator= ( const Date& date )
{
	if( &date != this )
	{
		_lYear = date._lYear;
		_lMonth = date._lMonth;
		_lDay = date._lDay;
	}

	return *this;
}

Date::edayOfWeek Date::dayOfWeek() const
{
	struct tm tm;
	memset( &tm, 0, sizeof( tm ) );
	tm.tm_year = _lYear - 1900;
	tm.tm_mon = _lMonth - 1;
	tm.tm_mday = _lDay;
	if( mktime( &tm ) == -1 )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	return (edayOfWeek)tm.tm_wday;
}

long Date::dayOfYear() const
{
	long lRtn = 0;

	for( int nIndex = 0; nIndex < ( _lMonth - 1 ); nIndex++ )
	{
		if( lMonthDayS[nIndex] == -1 )
		{
			if( IS_LEAP_YEAR( _lYear ) )
				lRtn += 29;
			else
				lRtn += 28;
		}
		else
		{
			lRtn += lMonthDayS[nIndex];
		}
	}
	lRtn += _lDay;

/*	struct tm tm;
	memset( &tm, 0, sizeof( tm ) );
	tm.tm_year = _lYear - 1900;
	tm.tm_mon = _lMonth - 1;
	tm.tm_mday = _lDay;
	if( mktime( &tm ) == -1 )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	return tm.tm_yday; */

	return lRtn;
}

long Date::lastDay() const
{
	long lRtn = lMonthDayS[( _lMonth - 1 )];

	if( lRtn == -1 )
	{
		if( IS_LEAP_YEAR( _lYear ) )
			lRtn = 29;
		else
			lRtn = 28;
	}

	return lRtn;
}

void Date::year( long lYear )
{
	if( ( lYear < 1790 ) || ( lYear > 2037 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_lYear = lYear;
}

void Date::month( long lMonth )
{
	if( ( lMonth > 12 ) || ( lMonth < 1 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_lMonth = lMonth;
}

void Date::day( long lDay )
{
	if( _lYear == 0 )
	{
		DateTimeException e( DTE_INVALID_YEAR );
		throw e;
	}
	if( _lMonth == 0 )
	{
		DateTimeException e( DTE_INVALID_MONTH );
		throw e;
	}

	long lMaxDays = lMonthDayS[( _lMonth - 1 )];
	if( lMaxDays == -1 )
	{
		if( IS_LEAP_YEAR( _lYear ) )
			lMaxDays = 29;
		else
			lMaxDays = 28;
	}

	if( ( lDay > lMaxDays ) || ( lDay < 1 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_lDay = lDay;
}

String Date::format( const String& strFormat ) const
{
	long lYearLength = strFormat.mask( strFormat.createMask( pszMaskYear ) ).trim().length();
	long lMonthLength = strFormat.mask( strFormat.createMask( pszMaskMonth ) ).trim().length();
	long lDayLength = strFormat.mask( strFormat.createMask( pszMaskDay ) ).trim().length();
	long lWeekdayLength = strFormat.mask( strFormat.createMask( pszMaskWeekday ) ).trim().length();
	String strYear;
	String strMonth;
	String strDay;
	String strWeekday;

	if( lYearLength == 0 )
		;
	else if( lYearLength == 2 )
	{
		strYear.format( "%02i", _lYear % 100 );
	}
	else if( lYearLength == 4 )
	{
		strYear.format( "%04i", _lYear );
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	if( lMonthLength == 0 )
		;
	else if( lMonthLength == 1 )
	{
		strMonth.format( "%i", _lMonth );
	}
	else if( lMonthLength == 2 )
	{
		strMonth.format( "%02i", _lMonth );
	}
	else if( lMonthLength == 3 )
	{
		strMonth.format( "%s", pszMonthS[( _lMonth - 1 )] );
	}
	else if( lMonthLength == 4 )
	{
		strMonth.format( "%s", pszLongMonthS[( _lMonth - 1 )] );
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	if( lDayLength == 0 )
		;
	else if( lDayLength == 1 )
	{
		strDay.format( "%i", _lDay );
	}
	else if( lDayLength == 2 )
	{
		strDay.format( "%02i", _lDay );
	}
	else if( lDayLength == 3 )
	{
		strDay.format( "%03i", dayOfYear() );
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	if( lWeekdayLength == 0 )
		;
	else if( lWeekdayLength == 1 )
	{
		strWeekday = pszDayS[dayOfWeek()];
	}
	else if( lWeekdayLength == 2 )
	{
		strWeekday = pszLongDayS[dayOfWeek()];
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	String strRtn( strFormat );

	if( strFormat.createMask( pszMaskYear ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskYear ), strYear );
	}
	if( strFormat.createMask( pszMaskMonth ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskMonth ), strMonth );
	}
	if( strFormat.createMask( pszMaskDay ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskDay ), strDay );
	}
	if( strFormat.createMask( pszMaskWeekday ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskWeekday ), strWeekday );
	}

	return strRtn;
}

Date Date::now()
{
	time_t t( time( NULL ) );
	struct tm* ptm = localtime( &t );

	return Date( ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday );
}

void Date::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl << (const char*)format() << std::endl;
}

void Date::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n" << (const char*)format() << "\n";
}
