/*
	$Id: Profile.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Profile.hpp>

#include <librfc/String.hpp>

#include <librfc/Socket.hpp>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

USE_RFC_NAMESPACE

Profile::Mode Profile::_mode;

//////////////////////////////////////////////////////////////////////////
//
// ProfileEntry
//

Profile::ProfileEntry::ProfileEntry( const char* pszName, const char* pszValue )
			: _pszName( NULL ),
			_pszValue( NULL ),
			_bModified( false ),
			_bGlobal( false ),
			_pnext( NULL )
{
	if( pszName != NULL )
	{
		_pszName = new char[( strlen( pszName ) + 1 )];
		strcpy( _pszName, pszName );
	}

	if( pszValue != NULL )
	{
		_pszValue = new char[( strlen( pszValue ) + 1 )];
		strcpy( _pszValue, pszValue );
	}

	switch( Profile::mode() )
	{
		case Profile::running:
			_bModified = true;
			break;
		case Profile::reading_global:
			_bGlobal = true;
			break;
		default:
			break;
	}
}

Profile::ProfileEntry::ProfileEntry( const char* pszName, long lValue )
			: _pszName( NULL ),
			_pszValue( NULL ),
			_bModified( false ),
			_bGlobal( false ),
			_pnext( NULL )
{
	if( pszName != NULL )
	{
		_pszName = new char[( strlen( pszName ) + 1 )];
		strcpy( _pszName, pszName );
	}

	_pszValue = new char[16];
	sprintf( _pszValue, "%li", lValue );

	switch( Profile::mode() )
	{
		case Profile::running:
			_bModified = true;
			break;
		case Profile::reading_global:
			_bGlobal = true;
			break;
		default:
			break;
	}
}

Profile::ProfileEntry::~ProfileEntry()
{
	if( _pszName != NULL )
		delete _pszName;
	if( _pszValue != NULL )
		delete _pszValue;
	if( _pnext != NULL )
		delete _pnext;
}

void Profile::ProfileEntry::write( FILE* h )
{
	if( ( _pszName != NULL ) &&
		( _pszValue != NULL ) &&
		( !_bGlobal ) )
	{
		fprintf( h, "%s=%s\n", _pszName, _pszValue );
	}
}

const char* Profile::ProfileEntry::name() const
{
	return _pszName;
}

const char* Profile::ProfileEntry::valueString() const
{
	return _pszValue;
}

long Profile::ProfileEntry::valueLong() const
{
	return atol( _pszValue );
}

void Profile::ProfileEntry::setValue( const char* pszValue )
{
	if( _pszValue != NULL )
		delete _pszValue;

	_pszValue = new char[( strlen( pszValue ) + 1 )];
	strcpy( _pszValue, pszValue );

	_bModified = true;
	_bGlobal = false;
}

void Profile::ProfileEntry::setValue( long lValue )
{
	if( _pszValue != NULL )
		delete _pszValue;

	_pszValue = new char[16];
	sprintf( _pszValue, "%li", lValue );

	_bModified = true;
	_bGlobal = false;
}


//////////////////////////////////////////////////////////////////////////
//
// ProfileSection
//

Profile::ProfileSection::ProfileSection( const char* pszName )
				: _pfirst( NULL ),
				_plast( NULL ),
				_pnext( NULL ),
				_pszName( NULL )
{
	if( pszName != NULL )
	{
		_pszName = new char[( strlen( pszName ) + 1 )];
		strcpy( _pszName, pszName );
	}
}

Profile::ProfileSection::~ProfileSection()
{
	if( _pszName != NULL )
		delete _pszName;
	if( _pfirst != NULL )
		delete _pfirst;
	if( _pnext != NULL )
		delete _pnext;
}

void Profile::ProfileSection::write( FILE* h )
{
	if( _pszName != NULL )
	{
		fprintf( h, "[%s]\n", _pszName );
	}

	ProfileEntry* pentry = _pfirst;
	while( pentry != NULL )
	{
		pentry->write( h );
		pentry = pentry->_pnext;
	}
}

const char* Profile::ProfileSection::entryString( const char* pszEntryName, const char* pszDefault ) const
{
	ProfileEntry* pentry = find( pszEntryName );
	if( pentry != NULL )
		return pentry->valueString();
	else
		return pszDefault;
}

long Profile::ProfileSection::entryLong( const char* pszEntryName, long lDefault ) const
{
	ProfileEntry* pentry = find( pszEntryName );

	if( pentry != NULL )
		return pentry->valueLong();
	else
		return lDefault;
}

void Profile::ProfileSection::setEntryValue( const char* pszEntryName, const char* pszValue )
{
	ProfileEntry* pentry = find( pszEntryName );
	if( pentry != NULL )
		pentry->setValue( pszValue );
	else
		addEntry( pszEntryName, pszValue );
}

void Profile::ProfileSection::setEntryValue( const char* pszEntryName, long lValue )
{
	ProfileEntry* pentry = find( pszEntryName );
	if( pentry != NULL )
		pentry->setValue( lValue );
	else
		addEntry( pszEntryName, lValue );
}

Profile::ProfileEntry* Profile::ProfileSection::find( const char* pszEntryName ) const
{
	ProfileEntry* pcurrent = _pfirst;
	ProfileEntry* pfound = NULL;

	while( ( pcurrent != NULL ) &&
		( pfound == NULL ) )
	{
		if( !strcmp( pszEntryName, pcurrent->name() ) )
			pfound = pcurrent;
		else
			pcurrent = pcurrent->_pnext;
	}

	return pfound;
}

void Profile::ProfileSection::addEntry( const char* pszEntryName, const char* pszValue )
{
	ProfileEntry* pentry = new ProfileEntry( pszEntryName, pszValue );
	if( _pfirst == NULL )
		_pfirst = _plast = pentry;
	else
		_plast = _plast->_pnext = pentry;
}

void Profile::ProfileSection::addEntry( const char* pszEntryName, long lValue )
{
	ProfileEntry* pentry = new ProfileEntry( pszEntryName, lValue );
	if( _pfirst == NULL )
		_pfirst = _plast = pentry;
	else
		_plast = _plast->_pnext = pentry;
}


//////////////////////////////////////////////////////////////////////////
//
// Profile
//

Profile::Profile( const char* pszProfileLocal, const char* pszProfileGlobal )
		: _pfirst( NULL ),
		_plast( NULL ),
		_pszProfileLocal( NULL ),
		_pszProfileGlobal( NULL ),
		_bProfileFound( false )
{
	if( pszProfileLocal != NULL )
	{
		String strProfileLocal;
		_pszProfileLocal = new char[( strlen( pszProfileLocal ) + 1 )];
		strcpy( _pszProfileLocal, pszProfileLocal );

#if defined (_WINDOWS)
		strProfileLocal.format( "%s\\%s", getenv( "HOMEPATH" ), _pszProfileLocal );
		FILE* hLocal = fopen( strProfileLocal, "rt" );
#elif defined (LINUX)
		strProfileLocal.format( "%s/%s", getenv( "HOME" ), _pszProfileLocal );
		FILE* hLocal = fopen( strProfileLocal, "r" );
#endif

		if( hLocal != NULL )
		{
			_mode = reading_local;
			processFile( hLocal );
			fclose( hLocal );
			_bProfileFound = true;
		}
	}

	if( pszProfileGlobal != NULL )
	{
		String strProfileGlobal;
		_pszProfileGlobal = new char[( strlen( pszProfileGlobal ) + 1 )];
		strcpy( _pszProfileGlobal, pszProfileGlobal );

#if defined (_WINDOWS)
		strProfileGlobal.format( "c:\\etc\\%s", _pszProfileGlobal );
		FILE* hGlobal = fopen( strProfileGlobal, "rt" );
#elif defined (LINUX)
		strProfileGlobal.format( "/etc/%s", _pszProfileGlobal );
		FILE* hGlobal = fopen( strProfileGlobal, "r" );
#endif

		if( hGlobal != NULL )
		{
			_mode = reading_global;
			processFile( hGlobal );
			fclose( hGlobal );
			_bProfileFound = true;
		}
	}

	_mode = running;
}

Profile::~Profile()
{
	_mode = writing;

	if( _pszProfileLocal != NULL )
	{
		char szProfileLocal[128];
		sprintf( szProfileLocal, "%s/%s", getenv( "HOME" ), _pszProfileLocal );
		FILE* hLocal = fopen( szProfileLocal, "w" );

		if( hLocal != NULL )
		{
			ProfileSection* psection = _pfirst;
			while( psection != NULL )
			{
				psection->write( hLocal );
				psection = psection->_pnext;
			}
			fclose( hLocal );
		}
		delete _pszProfileLocal;
		_pszProfileLocal = NULL;
	}
	if( _pszProfileGlobal != NULL )
	{
		delete _pszProfileGlobal;
		_pszProfileGlobal = NULL;
	}
}

const char* Profile::getProfileString( const char* pszSection, const char* pszEntry, const char* pszDefault )
{
	ProfileSection* psection = find( pszSection );
	if( psection == NULL )
		return pszDefault;

	return psection->entryString( pszEntry, pszDefault );
}

long Profile::getProfileLong( const char* pszSection, const char* pszEntry, long lDefault )
{
	ProfileSection* psection = find( pszSection );
	if( psection == NULL )
		return lDefault;

	return psection->entryLong( pszEntry, lDefault );
}

void Profile::writeProfileString( const char* pszSection, const char* pszEntry, const char* pszValue )
{
	ProfileSection* psection = find( pszSection );
	if( psection == NULL )
		psection = addSection( pszSection );

	psection->setEntryValue( pszEntry, pszValue );
}

void Profile::writeProfileLong( const char* pszSection, const char* pszEntry, long lValue )
{
	ProfileSection* psection = find( pszSection );
	if( psection == NULL )
		psection = addSection( pszSection );

	psection->setEntryValue( pszEntry, lValue );
}

Profile::Mode Profile::mode()
{
	return _mode;
}

bool Profile::profileFound() const
{
	return _bProfileFound;
}

Profile::ProfileSection* Profile::addSection( const char* pszName )
{
	ProfileSection* psection = new ProfileSection( pszName );
	if( _pfirst == NULL )
		_pfirst = _plast = psection;
	else
		_plast = _plast->_pnext = psection;

	return psection;
}

Profile::ProfileSection* Profile::find( const char* pszName ) const
{
	ProfileSection* pcurrent = _pfirst;
	ProfileSection* pfound = NULL;

	while( ( pcurrent != NULL ) &&
		( pfound == NULL ) )
	{
		if( !strcmp( pszName, pcurrent->_pszName ) )
			pfound = pcurrent;
		else
			pcurrent = pcurrent->_pnext;
	}

	return pfound;
}

void Profile::processFile( FILE* h )
{
	char szBuffer[512];
	ProfileSection* psection = NULL;

	while( !feof( h ) )
	{
		fgets( szBuffer, sizeof szBuffer, h );
		if( szBuffer[0] == '[' )
		{
			char* pszStart = (char*)szBuffer;
			pszStart++;
			char* pszEnd = strchr( pszStart, ']' );
			if( pszEnd != NULL )
				*pszEnd = 0;

			psection = find( pszStart );
			if( psection == NULL )
				psection = addSection( pszStart );
		}
		else
		{
			if( psection != NULL )
			{
				char* pszValue = strchr( szBuffer, '=' );
				if( pszValue != NULL )
				{
					*pszValue = 0;
					pszValue++;
					char* pszEnd = strchr( pszValue, '\n' );
					if( pszEnd != NULL )
						*pszEnd = 0;
					psection->addEntry( szBuffer, pszValue );
				}
			}
		}
	}
}

void Profile::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl
		<< "L=" << ( _pszProfileLocal == NULL ? "(null)" : _pszProfileLocal )
		<< " G=" << ( _pszProfileGlobal == NULL ? "(null)" : _pszProfileGlobal )
		<< std::endl;
}

void Profile::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n"
		<< "L=" << ( _pszProfileLocal == NULL ? "(null)" : _pszProfileLocal )
		<< " G=" << ( _pszProfileGlobal == NULL ? "(null)" : _pszProfileGlobal )
		<< "\n";
}
