/*
	$Id: Socket.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Socket.hpp>
#include <librfc/SocketException.hpp>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#if defined (LINUX)
#	include <unistd.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#	include <netdb.h>
#	include <sys/poll.h>
#	include <sys/socket.h>
#	include <sys/ioctl.h>
#	include <sys/un.h>
#	include <errno.h>
#elif defined (_WINDOWS)
#	include <winsock2.h>
#endif


USE_RFC_NAMESPACE

Socket::Socket( Domain domain/*= Inet*/ )
	: _domain( domain )
{
	if( _domain == Inet )
		_s = ::socket( AF_INET, SOCK_STREAM, IPPROTO_TCP );
	else if( _domain == Unix )
		_s = ::socket( PF_UNIX, SOCK_STREAM, 0 );

	if( _s == -1 )
	{
		throw SocketException( h_errno );
	}

	_strWaiting = "";
	_strWaiting.excludeFromDump();
}

Socket::Socket( const Socket& skt )
{
	_s = skt._s;
	_domain = skt._domain;

	_strWaiting = "";
	_strWaiting.excludeFromDump();
}

Socket::Socket( int s )
{
	_s = s;

	_strWaiting = "";
	_strWaiting.excludeFromDump();
}

Socket::~Socket()
{
}

const Socket& Socket::operator= ( const Socket& skt )
{
	if( this == & skt )
		return *this;

	_s = skt._s;
	_domain = skt._domain;

	_strWaiting = skt._strWaiting;

	return *this;
}

bool Socket::bind( short nPort )
{
	int nRC;

	if( _domain == Inet )
	{
		struct sockaddr_in addr;

		addr.sin_family = AF_INET;
		addr.sin_port = htons( nPort );
		addr.sin_addr.s_addr = INADDR_ANY;

		nRC = ::bind( _s, (const struct sockaddr*)&addr, sizeof addr  );
	}
	else
	{
		SocketException e( SXN_WRONGDOMAIN );
		throw e;
	}

	if( nRC == -1 )
	{
		close();

		throw SocketException( h_errno );
	}

	return true;
}

bool Socket::bind( const String& strFile )
{
#ifdef LINUX
	int nRC;

	if( _domain == Unix )
	{
		struct sockaddr_un addr;

		addr.sun_family = AF_UNIX;
		strcpy( addr.sun_path, strFile );

		nRC = ::bind( _s, (const struct sockaddr*)&addr, sizeof addr  );
	}
	else
	{
		SocketException e( SXN_WRONGDOMAIN );
		throw e;
	}

	if( nRC == -1 )
	{
		close();

		throw SocketException( h_errno );
	}
#endif
	return true;
}

bool Socket::listen()
{
	int nRC;

	nRC = ::listen( _s, SOMAXCONN );

	if( nRC == -1 )
	{
		close();

		throw SocketException( h_errno );
	}

	return true;
}

Socket Socket::accept()
{
	int nRC;

	nRC = ::accept( _s, NULL, NULL );
	if( nRC == -1 )
	{
		throw SocketException( h_errno );
	}

	return Socket( nRC );
}

bool Socket::connect( const String& strHost, short nPort )
{
	long nIpAddr;

	if( atoi( strHost ) )
	{
		nIpAddr = inet_addr( strHost );
	}
	else
	{
		struct hostent* phostent;
		phostent = gethostbyname( strHost );

		if( phostent == NULL )
		{
			throw SocketException( h_errno );
		}

		char szAddrlist[5];

		memcpy(szAddrlist, phostent->h_addr_list[0], 4);
		nIpAddr = *(unsigned long*)&szAddrlist;
	}

	struct sockaddr_in addr;

	addr.sin_family = AF_INET;
	addr.sin_port = htons( nPort );
	addr.sin_addr.s_addr = nIpAddr;

	if( ::connect( _s, (const struct sockaddr*)&addr, sizeof( addr ) )
		== -1 )
	{
		close();

		throw SocketException( h_errno );
	}

	return true;
}

bool Socket::connect( const String& strFile )
{
#ifdef LINUX
	struct sockaddr_un addr;

	addr.sun_family = AF_UNIX;
	strcpy( addr.sun_path, strFile );

	if( ::connect( _s, (const struct sockaddr*)&addr, sizeof( addr ) )
		== -1 )
	{
		close();

		throw SocketException( h_errno );
	}
#endif
	return true;
}

unsigned long Socket::send( const String& strData )
{
	return ::send( _s, strData, strData.length(), 0 );
}

unsigned long Socket::recv( String& strData, unsigned long ulTimeout )
{
	unsigned long ulWaiting = 0;
	long lRC;

	if( _s == -1 )
		throw SocketException( -1 );

	fd_set sktsRead;
	fd_set sktsExcept;

	FD_ZERO( &sktsRead );
	FD_ZERO( &sktsExcept );
	FD_SET( _s, &sktsRead );
	FD_SET( _s, &sktsExcept );

	if( ulTimeout == (unsigned long)-1 )
		lRC = select( _s + 1, &sktsRead, NULL, &sktsExcept, NULL );
	else
	{
		struct timeval tv;
		tv.tv_usec = ulTimeout % 1000000;
		tv.tv_sec = ulTimeout / 1000000;
		lRC = select( _s + 1, &sktsRead, NULL, &sktsExcept, &tv );
		if( !lRC )
			return 0;
	}

	if( lRC == -1 )
	{
		throw SocketException( h_errno );
	}

	if( FD_ISSET( _s, &sktsExcept ) )
	{
		throw SocketException( h_errno );
	}

#if defined (LINUX)
	if( ioctl( _s, FIONREAD, &ulWaiting ) == -1 )
#elif defined (_WINDOWS)
	if( ioctlsocket( _s, FIONREAD, &ulWaiting ) == -1 )
#endif
	{
		close();

		throw SocketException( h_errno );
	}
	else
	{
		if( ulWaiting )
		{
			char* pszBuffer = new char[( ulWaiting + 1 )];
			pszBuffer[ulWaiting] = '\0';
			::recv( _s, pszBuffer, ulWaiting, 0 );
			strData = pszBuffer;
			delete pszBuffer;
		}
		else
		{
			close();

			throw SocketException( h_errno );
		}
	}

	return ulWaiting;
}

void Socket::close()
{
	if( _s != -1 )
	{
		::shutdown( _s, 1 );
		::shutdown( _s, 2 );
#if defined (LINUX)
		::close( _s );
#elif defined (_WINDOWS)
		::closesocket( _s );
#endif
		_s = -1;
	}
}

const char* Socket::resolveSocketError( unsigned long ulError )
{
#if defined (LINUX)
	switch( ulError )
	{
		case ENOMEM:
			return "ENOMEM";
		case EPIPE:
			return "EPIPE";
//		case EPERM:
//			return "EPERM";
		case EINTR:
			return "EINTR";
		case EBADF:
			return "EBADF";
		case EACCES:
			return "EACCES";
		case EFAULT:
			return "EFAULT";
		case EINVAL:
			return "EINVAL";
		case EMFILE:
			return "EMFILE";
		case EWOULDBLOCK:
			return "EWOULDBLOCK";
		case EINPROGRESS:
			return "EINPROGRESS";
		case EALREADY:
			return "EALREADY";
		case ENOTSOCK:
			return "ENOTSOCK";
		case EDESTADDRREQ:
			return "EDESTADDRREQ";
		case EMSGSIZE:
			return "EMSGSIZE";
		case EPROTOTYPE:
			return "EPROTOTYPE";
		case ENOPROTOOPT:
			return "ENOPROTOOPT";
		case EPROTONOSUPPORT:
			return "EPROTONOSUPPORT";
		case ESOCKTNOSUPPORT:
			return "ESOCKTNOSUPPORT";
		case EOPNOTSUPP:
			return "EOPNOTSUPP";
		case EPFNOSUPPORT:
			return "EPFNOSUPPORT";
		case EAFNOSUPPORT:
			return "EAFNOSUPPORT";
		case EADDRINUSE:
			return "EADDRINUSE";
		case EADDRNOTAVAIL:
			return "EADDRNOTAVAIL";
		case ENETDOWN:
			return "ENETDOWN";
		case ENETUNREACH:
			return "ENETUNREACH";
		case ENETRESET:
			return "ENETRESET";
		case ECONNABORTED:
			return "ECONNABORTED";
		case ECONNRESET:
			return "ECONNRESET";
		case ENOBUFS:
			return "ENOBUFS";
		case EISCONN:
			return "EISCONN";
		case ENOTCONN:
			return "ENOTCONN";
		case ESHUTDOWN:
			return "ESHUTDOWN";
		case ETOOMANYREFS:
			return "ETOOMANYREFS";
		case ETIMEDOUT:
			return "ETIMEDOUT";
		case ECONNREFUSED:
			return "ECONNREFUSED";
		case ELOOP:
			return "ELOOP";
		case ENAMETOOLONG:
			return "ENAMETOOLONG";
		case EHOSTDOWN:
			return "EHOSTDOWN";
		case EHOSTUNREACH:
			return "EHOSTUNREACH";
		case ENOTEMPTY:
			return "ENOTEMPTY";
		case EUSERS:
			return "EUSERS";
		case EDQUOT:
			return "EDQUOT";
		case ESTALE:
			return "ESTALE";
		case EREMOTE:
			return "EREMOTE";
		case HOST_NOT_FOUND:
			return "HOST_NOT_FOUND/EPERM";
		case TRY_AGAIN:
			return "TRY_AGAIN";
		case NO_RECOVERY:
			return "NO_RECOVERY";
//		case NO_DATA:					same as EINTR
//			return "NO_DATA";
//		case NO_ADDRESS:			same as NO_DATA
//			return "NO_ADDRESS";
		default:
			return "Error not found";
	}
#elif defined (_WINDOWS)
	switch( ulError )
	{
		case WSAEINTR:
			return "EINTR";
		case WSAEBADF:
			return "EBADF";
		case WSAEACCES:
			return "EACCES";
		case WSAEFAULT:
			return "EFAULT";
		case WSAEINVAL:
			return "EINVAL";
		case WSAEMFILE:
			return "EMFILE";
		case WSAEWOULDBLOCK:
			return "EWOULDBLOCK";
		case WSAEINPROGRESS:
			return "EINPROGRESS";
		case WSAEALREADY:
			return "EALREADY";
		case WSAENOTSOCK:
			return "ENOTSOCK";
		case WSAEDESTADDRREQ:
			return "EDESTADDRREQ";
		case WSAEMSGSIZE:
			return "EMSGSIZE";
		case WSAEPROTOTYPE:
			return "EPROTOTYPE";
		case WSAENOPROTOOPT:
			return "ENOPROTOOPT";
		case WSAEPROTONOSUPPORT:
			return "EPROTONOSUPPORT";
		case WSAESOCKTNOSUPPORT:
			return "ESOCKTNOSUPPORT";
		case WSAEOPNOTSUPP:
			return "EOPNOTSUPP";
		case WSAEPFNOSUPPORT:
			return "EPFNOSUPPORT";
		case WSAEAFNOSUPPORT:
			return "EAFNOSUPPORT";
		case WSAEADDRINUSE:
			return "EADDRINUSE";
		case WSAEADDRNOTAVAIL:
			return "EADDRNOTAVAIL";
		case WSAENETDOWN:
			return "ENETDOWN";
		case WSAENETUNREACH:
			return "ENETUNREACH";
		case WSAENETRESET:
			return "ENETRESET";
		case WSAECONNABORTED:
			return "ECONNABORTED";
		case WSAECONNRESET:
			return "ECONNRESET";
		case WSAENOBUFS:
			return "ENOBUFS";
		case WSAEISCONN:
			return "EISCONN";
		case WSAENOTCONN:
			return "ENOTCONN";
		case WSAESHUTDOWN:
			return "ESHUTDOWN";
		case WSAETOOMANYREFS:
			return "ETOOMANYREFS";
		case WSAETIMEDOUT:
			return "ETIMEDOUT";
		case WSAECONNREFUSED:
			return "ECONNREFUSED";
		case WSAELOOP:
			return "ELOOP";
		case WSAENAMETOOLONG:
			return "ENAMETOOLONG";
		case WSAEHOSTDOWN:
			return "EHOSTDOWN";
		case WSAEHOSTUNREACH:
			return "EHOSTUNREACH";
		case WSAENOTEMPTY:
			return "ENOTEMPTY";
		case WSAEUSERS:
			return "EUSERS";
		case WSAEDQUOT:
			return "EDQUOT";
		case WSAESTALE:
			return "ESTALE";
		case WSAEREMOTE:
			return "EREMOTE";
		case WSAEPROCLIM:
			return "EPROCLIM";
		case WSAEDISCON:
			return "WSAEDISCON";
		case WSASYSNOTREADY:
			return "WSASYSNOTREADY";
		case WSAVERNOTSUPPORTED:
			return "WSAVERNOTSUPPORTED";
		case WSANOTINITIALISED:
			return "WSANOTINITIALISED";
		case WSAHOST_NOT_FOUND:
			return "HOST_NOT_FOUND";
		case WSATRY_AGAIN:
			return "TRY_AGAIN";
		case WSANO_RECOVERY:
			return "NO_RECOVERY";
//		case NO_DATA:					same as EINTR
//			return "NO_DATA";
//		case NO_ADDRESS:			same as NO_DATA
//			return "NO_ADDRESS";
		default:
			return "Error not found";
	}
#endif
}

void Socket::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << "\n";
}

void Socket::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n";
}

Socket& Socket::operator << ( const char* psz )
{
	_strWaiting += psz;

	return *this;
}

Socket& Socket::operator << ( short n )
{
	char sz[64];
	sprintf( sz, "%i", n );
	_strWaiting += sz;

	return *this;
}

Socket& Socket::operator << ( unsigned short u )
{
	char sz[64];
	sprintf( sz, "%u", u );
	_strWaiting += sz;

	return *this;
}

Socket& Socket::operator << ( long l )
{
	char sz[64];
	sprintf( sz, "%li", l );
	_strWaiting += sz;

	return *this;
}

Socket& Socket::operator << ( unsigned long ul )
{
	char sz[64];
	sprintf( sz, "%lu", ul );
	_strWaiting += sz;

	return *this;
}

Socket& Socket::operator << ( Action a )
{
	if( a == xmit )
	{
		send( _strWaiting );
		_strWaiting = "";
	}

	return *this;
}
