/*
	$Id: Time.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/Time.hpp>
#include <librfc/DateTimeException.hpp>
#include <time.h>

#include <librfc/Socket.hpp>

USE_RFC_NAMESPACE

const char* pszTimeLegalDelimiters = " -/\\.:";

const char* pszMaskHour = "H";
const char* pszMaskHour12 = "h";
const char* pszMaskMinute = "N";
const char* pszMaskSecond = "S";
const char* pszMaskSecondDot = "S.";
const char* pszMaskAmPm = "P";

const char* pszAMPM[] = { "AM", "PM" };

Time::Time()
{
	_strFormat = GetFCCString( SE_DEFAULT_TIME_FORMAT );
	_lHour = 0;
	_lMinute = 0;
	_dSecond = 0.0;
}

Time::Time( long lHour, long lMinute, long lSecond )
{
	_strFormat = GetFCCString( SE_DEFAULT_TIME_FORMAT );
	hour( lHour );
	minute( lMinute );
	second( lSecond );
}

Time::Time( long lHour, long lMinute, double dSecond )
{
	_strFormat = GetFCCString( SE_DEFAULT_TIME_FORMAT );
	hour( lHour );
	minute( lMinute );
	second( dSecond );
}

Time::Time( const String& strTime, const String& strFormat )
{
	_strFormat = strFormat;

	unsigned long ulMaskLength = _strFormat.length();

	if( strTime.length() != ulMaskLength )
	{
		DateTimeException e( DTE_MASK_MISMATCH );
		throw e;
	}

	for( unsigned long ulIndex = 0; ulIndex < ulMaskLength; ulIndex++ )
	{
		if( strchr( pszTimeLegalDelimiters, _strFormat.at( ulIndex ) ) != NULL )
		{
			if( _strFormat.at( ulIndex ) != strTime.at( ulIndex ) )
			{
				DateTimeException e( DTE_MASK_MISMATCH );
				throw e;
			}
		}
	}

	hour( (long)strTime.mask( _strFormat.createMask( pszMaskHour ) ).trim() );
	minute( (long)strTime.mask( _strFormat.createMask( pszMaskMinute ) ).trim() );
	second( (double)strTime.mask( _strFormat.createMask( pszMaskSecondDot ) ).trim() );
}

Time::Time( const Time& time )
{
	_strFormat = time._strFormat;
	_lHour = time._lHour;
	_lMinute = time._lMinute;
	_dSecond = time._dSecond;
}

Time::~Time()
{
}

const Time& Time::operator= ( const String& strTime )
{
	*this = Time( strTime, _strFormat );

	return *this;
}

const Time& Time::operator= ( const Time& time )
{
	if( &time != this )
	{
		_lHour = time._lHour;
		_lMinute = time._lMinute;
		_dSecond = time._dSecond;
	}

	return *this;
}

void Time::hour( long lHour )
{
	if( ( lHour > 23 ) || ( lHour < 0 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_lHour = lHour;
}

void Time::minute( long lMinute )
{
	if( ( lMinute > 59 ) || ( lMinute < 0 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_lMinute = lMinute;
}

void Time::second( long lSecond )
{
	if( ( lSecond > 59 ) || ( lSecond < 0 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_dSecond = lSecond;
}

void Time::second( double dSecond )
{
	if( ( dSecond >= 60 ) || ( dSecond < 0 ) )
	{
		DateTimeException e( DTE_INVALID_VALUE );
		throw e;
	}

	_dSecond = dSecond;
}

String Time::format( const String& strFormat ) const
{
	bool b12( false );
	
	long lHourLength = strFormat.mask( strFormat.createMask( pszMaskHour ) ).trim().length();
	if( !lHourLength )
	{
		b12 = true;
		lHourLength  = strFormat.mask( strFormat.createMask( pszMaskHour12 ) ).trim().length();
	}
	long lMinuteLength = strFormat.mask( strFormat.createMask( pszMaskMinute ) ).trim().length();
	long lSecondLength = strFormat.mask( strFormat.createMask( pszMaskSecond ) ).trim().length();
	long lAmPmLength = strFormat.mask( strFormat.createMask( pszMaskAmPm ) ).trim().length();
	String strHour;
	String strMinute;
	String strSecond;
	String strAmPm;

	if( lHourLength == 0 )
		;
	else if( lHourLength == 1 )
	{
		if( b12 )
			strHour.format( "%i", _lHour % 12 );
		else
			strHour.format( "%i", _lHour );
	}
	else if( lHourLength == 2 )
	{
		if( b12 )
			strHour.format( "%02i", _lHour % 12 );
		else
			strHour.format( "%02i", _lHour );
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	if( lMinuteLength == 0 )
		;
	else if( lMinuteLength == 1 )
	{
		strMinute.format( "%i", _lMinute );
	}
	else if( lMinuteLength == 2 )
	{
		strMinute.format( "%02i", _lMinute );
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	if( lSecondLength == 0 )
		;
	else if( lSecondLength == 1 )
	{
		strSecond.format( "%i", (long)_dSecond );
	}
	else if( lSecondLength == 2 )
	{
		strSecond.format( "%02i", (long)_dSecond );
	}
	else if( ( lSecondLength == 3 ) || ( lSecondLength > 8 ) )
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}
	else
	{
		String strSecondFormat;
		strSecondFormat.format( "%%0%i.%if", lSecondLength, lSecondLength - 3 );
		strSecond.format( strSecondFormat, _dSecond );
	}

	if( lAmPmLength == 0 )
		;
	else if( lAmPmLength == 1 )
	{
		strAmPm.format( "%c", pszAMPM[_lHour / 12][0] );
	}
	else if( lAmPmLength == 2 )
	{
		strAmPm = pszAMPM[_lHour/12];
	}
	else
	{
		DateTimeException e( DTE_INVALID_FORMAT );
		throw e;
	}

	String strRtn( strFormat );

	if( strFormat.createMask( pszMaskHour ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskHour ), strHour );
	}
	if( strFormat.createMask( pszMaskMinute ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskMinute ), strMinute );
	}
	if( strFormat.createMask( pszMaskSecondDot ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskSecondDot ), strSecond );
	}
	if( strFormat.createMask( pszMaskAmPm ).trim().length() )
	{
		strRtn = strRtn.replaceMask( strFormat.createMask( pszMaskAmPm ), strAmPm );
	}

	return strRtn;
}

Time Time::now( Where where )
{
	time_t t( time( NULL ) );

	struct tm* ptm;
	switch( where )
	{
		case local:
			ptm = localtime( &t );
			break;

		case utc:
			ptm = gmtime( &t );
			break;
	}

	return Time( ptm->tm_hour, ptm->tm_min, (long)ptm->tm_sec );
}

void Time::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl << (const char*)format() << std::endl;
}

void Time::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n" << (const char*)format() << "\n";
}
