/*
	$Id: XMLDoc.cpp 1205 2018-11-22 18:09:31Z john $

	This file is part of the Revelation Foundation Classes.

	The RFC is Copyright (c) 1989 - 1998, Revelation Software Ltd.
	The RFC is Copyright (c) 1998 - 2018, The 4c Company.

	The RFC is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <librfc/XMLDoc.hpp>

#include <string.h>
#include <stdio.h>

#ifdef USE_LIBXML
#	include <libxml/parser.h>
#	include <libxml/tree.h>
#else
#	include <sys/stat.h>
#	include <sys/types.h>
#	include <fcntl.h>
#endif

#ifdef LINUX
#	include <unistd.h>
#else
#	include <io.h>
#endif

#include <librfc/Socket.hpp>


USE_RFC_NAMESPACE

const char* XMLVersion = "1.0";

XMLDoc::XMLDoc()
	: _pnode( NULL ),
	_bIsNull( true )
{
}

XMLDoc::XMLDoc( const String& strRootNodeName, const String& strVersion, const String& strEncoding )
	: _pnode( NULL ),
	_bIsNull( false )
{
	_strVersion = strVersion;
	_strEncoding = strEncoding;
	_pnode = new XMLNode( NULL, strRootNodeName, String::null );
}

XMLDoc::XMLDoc( const String& strRootNodeName, bool bDocType, const String& strVersion, const String& strEncoding )
	: _pnode( NULL ),
	_bIsNull( false )
{
	_strVersion = strVersion;
	_strEncoding = strEncoding;

	if( bDocType )
	{
		_strDocType = strRootNodeName;
	}

	_pnode = new XMLNode( NULL, strRootNodeName, String::null );
}

XMLDoc::XMLDoc( const String& strDocType, const String& strVersion, const String& strEncoding, bool bCreateRootNode )
	: _pnode( NULL ),
	_bIsNull( true )
{
	_strVersion = strVersion;
	_strEncoding = strEncoding;
	_strDocType = strDocType;

	if( bCreateRootNode )
	{
		_pnode = new XMLNode( NULL, strDocType, String::null );
		_bIsNull = false;
	}
}

XMLDoc::XMLDoc( const char* pszStream )
	: _pnode( NULL ),
	_bIsNull( false )
{
	fromStream( pszStream );
}

#ifdef USE_LIBXML
XMLDoc::XMLDoc( const char* pszStream, int iSize )
	: _pnode( NULL ),
	_bIsNull( false )
{
	fromStream( pszStream, iSize );
}
#endif

XMLDoc::XMLDoc( const XMLDoc& doc )
	: _bIsNull( false )
{
	_strVersion = doc._strVersion;
	_strEncoding = doc._strEncoding;
	_strDocType = doc._strDocType;
	const XMLNode* pnodeRoot = doc.rootNode();
	if( pnodeRoot == NULL )
		_pnode = NULL;
	else
		_pnode = new XMLNode( NULL, *pnodeRoot );
	_bIsNull = doc._bIsNull;
}

XMLDoc::~XMLDoc()
{
	if( _pnode != NULL )
		delete _pnode;
}

const XMLDoc& XMLDoc::operator=( const XMLDoc& doc )
{
	if( &doc == this )
		return *this;

	_strVersion = doc._strVersion;
	_strEncoding = doc._strEncoding;
	_strDocType = doc._strDocType;
	const XMLNode* pnodeRoot = doc.rootNode();
	if( pnodeRoot == NULL )
		_pnode = NULL;
	else
		_pnode = new XMLNode( NULL, *pnodeRoot );
	_bIsNull = doc._bIsNull;

	return *this;
}

bool XMLDoc::isNull() const
{
	return _pnode == NULL ? true : false; //_bIsNull;
}

const XMLNode* XMLDoc::rootNode() const
{
	return _pnode;
}

XMLNode* XMLDoc::rootNode()
{
	return _pnode;
}

void XMLDoc::rootNode( const XMLNode& node )
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;

	String str = node.stream( false );
	const char* psz = str;
	_pnode = new XMLNode( NULL, psz );
	docType( _pnode->name() );
//	_strEncoding = "UTF-8";
}

void XMLDoc::rootNode( const String& strName, bool bDocType )
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;

	_pnode = new XMLNode( NULL, strName, String::null );
	if( bDocType )
		docType( _pnode->name() );
	_strEncoding = "UTF-8";
}

String XMLDoc::docType() const
{
	return _strDocType;
}

void XMLDoc::docType( const String& strDocType )
{
	_strDocType = strDocType;
}

String XMLDoc::version() const
{
	return _strVersion;
}

String XMLDoc::encoding() const
{
	return _strEncoding;
}

String XMLDoc::stream() const
{
	String strRtn;
	String strDocType;

	strRtn.format( "<?xml version=\"%s\" encoding=\"%s\"?>",
					(const char*)_strVersion, (const char*)_strEncoding );

	if( !_strDocType.isNull() )
	{
		strDocType.format( "<!DOCTYPE %s>", (const char*)_strDocType );
		strRtn += strDocType;
	}

	strRtn += ASC_NEWLINE;

	if( _pnode != NULL )
		strRtn += _pnode->stream();

	return strRtn;
}

#ifdef USE_LIBXML
void XMLDoc::create_elements( xmlNode *p_libxml_parent, XMLNode *pparent )
{
	xmlNode *p_libxml_node = p_libxml_parent;

	while( p_libxml_node != NULL )
	{
		xmlElementType type  = p_libxml_node->type;
		
		XMLNode *pnode = NULL;
		if( type == XML_ELEMENT_NODE )
		{
			String strName( (const char* ) p_libxml_node->name );
			String strNameSpace;

			xmlNs* pns = p_libxml_node->ns;
			if( NULL != pns )
			{
				strNameSpace = (const char*)pns->prefix;
			}
	
			if( strNameSpace.length() )
			{
				pnode = new XMLNode( pparent, strNameSpace, strName, String() );
			}
			else
			{
				pnode = new XMLNode( pparent, strName, String() );
			}

			xmlNode *p_libxml_child = p_libxml_node->children;
			//Only set value if single child of type text
			if( p_libxml_child != NULL && 
				p_libxml_child == p_libxml_node->last && //single child
				p_libxml_child->type == XML_TEXT_NODE )
			{
				String value( ( const char* ) p_libxml_child->content );
				pnode->value( value );
			}

			xmlAttr *p_libxml_attr = p_libxml_node->properties;
			while( p_libxml_attr != NULL )
			{
				if(  p_libxml_attr->type == XML_ATTRIBUTE_NODE )
				{
					String attr( (const char* ) p_libxml_attr->name );
					String value( ( const char* ) p_libxml_attr->children->content );
					pnode->attribute( attr, value );
				}
				p_libxml_attr = p_libxml_attr->next;
			}
		}

		if( pnode != NULL )
		{
			create_elements( p_libxml_node->children, pnode );
		}

		p_libxml_node = p_libxml_node->next;
	}
}

const XMLDoc& XMLDoc::fromStream( const char*& pszStream,
								  int iSize)
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;
	_pnode = NULL;

	if( pszStream == NULL )
	{
		throw XMLException( XE_D_INVALIDSTREAM );
	}


	xmlDocPtr doc; /* document treee */

	doc = xmlReadMemory( pszStream, iSize, "unknown.xml", NULL, 0 );
	
	if( doc == NULL )
	{
		/*xmlErrorPtr pPtr =*/ xmlGetLastError();

		throw XMLException( XE_D_MALFORMEDXML );
	}

	xmlNode *p_libxml_root = NULL;


	p_libxml_root = xmlDocGetRootElement( doc );

	_pnode = new XMLNode( NULL, "root", "");

	create_elements( p_libxml_root->children, _pnode );
	//copy root node attributes if any
	xmlAttr *p_libxml_attr = p_libxml_root->properties;
	while( p_libxml_attr != NULL )
	{
		if(  p_libxml_attr->type == XML_ATTRIBUTE_NODE )
		{
			String strAttr( (const char* ) p_libxml_attr->name );
			String strValue( ( const char* ) p_libxml_attr->children->content );
			_pnode->attribute( strAttr, strValue );
		}

		p_libxml_attr = p_libxml_attr->next;
	}

	if( _pnode == NULL )
	{
		_bIsNull = true;
	}

	xmlFreeDoc( doc );

	return *this;
}
#endif

const XMLDoc& XMLDoc::fromStream( const char*& pszStream )
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;
	_pnode = NULL;

	_strVersion = String::null;
	_strEncoding = String::null;
	_strDocType = String::null;

	if( pszStream == NULL )
	{
		throw XMLException( XE_D_INVALIDSTREAM );
	}

	XMLNode* pnode;

	try
	{
		while( pszStream > (const char*)2 )
		{
			pnode = new XMLNode( NULL, pszStream );
			switch( pnode->type() )
			{
				case XMLNode::typeComment:
					if( pnode->name() == XML_XML )
					{
						_strVersion = pnode->attribute( XML_VERSION );
						_strEncoding = pnode->attribute( XML_ENCODING );
					}
					delete pnode;
					break;

				case XMLNode::typeParameter:
					if( pnode->name() == XML_DOCTYPE )
					{
						XMLAttr* pattr = NULL;
						pattr = pnode->attribute( pattr );
						if( pattr != NULL )
							_strDocType = pattr->name();
					}
					delete pnode;
					break;

				case XMLNode::typeNormal:
					if( _pnode != NULL )
						delete pnode;
					else
						_pnode = pnode;
					break;

				default:
					delete pnode;
			}
		}
		if( _pnode == NULL )
			_bIsNull = true;
	}
	catch( const XMLException& e )
	{
		if( e.reason() != XE_N_NOTANODE )
			throw;
	}

	return *this;
}

const XMLDoc& XMLDoc::fromFile( const char *pszFileName )
#ifdef USE_LIBXML
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;
	_pnode = NULL;

	if( pszFileName == NULL )
	{
		throw XMLException( XE_D_INVALIDSTREAM );
	}


	xmlDocPtr doc; /* document treee */

	doc = xmlReadFile( pszFileName, NULL, 0 );
	
	if( doc == NULL )
	{
		/*xmlErrorPtr pPtr =*/ xmlGetLastError();

		throw XMLException( XE_D_MALFORMEDXML );
	}

	xmlNode *p_libxml_root = NULL;


	p_libxml_root = xmlDocGetRootElement( doc );

	_pnode = new XMLNode( NULL, "root", "");

	create_elements( p_libxml_root->children, _pnode );

	if( _pnode == NULL )
	{
		_bIsNull = true;
	}

	xmlFreeDoc( doc );

	return *this;
}
#else
{
	_bIsNull = false;

	if( _pnode != NULL )
		delete _pnode;
	_pnode = NULL;

	if( pszFileName == NULL )
	{
		throw XMLException( XE_D_INVALIDSTREAM );
	}

	struct stat buf;
	int rc = stat( pszFileName, &buf );
	if( -1 == rc )
	{
		throw XMLException( XE_D_NOSUCHFILE );
	}

	int nSize = buf.st_size;
	if( nSize < 5 )
	{
		throw XMLException( XE_D_INVALIDSTREAM );
	}
	
	int nH = open( pszFileName, O_RDONLY );
	if( -1 == nH )
	{
		throw XMLException( XE_D_NOSUCHFILE );
	}

	char* pszBuf = new char[nSize + 1];
	if( NULL == pszBuf )
	{
		throw XMLException( XE_D_OUTOFMEMORY );
	}

	char* pszDelete = pszBuf; // pszBuf will be modified by fromStream()

	memset( pszBuf, 0, nSize + 1 );

	int nBytesRead;
	int nBytesReadTotal( 0 );

	char* pszReadTo( pszBuf );

	try
	{
		while( nBytesReadTotal < nSize )
		{
			nBytesRead = read( nH, pszReadTo, nSize );
			nBytesReadTotal += nBytesRead;
			pszReadTo += nBytesRead;
		}
		
		fromStream( (const char*&)pszBuf );
	}
	catch( const XMLException& e )
	{
		close( nH );
		delete[] pszDelete;
		throw;
	}

	close( nH );
	delete[] pszDelete;

	return *this;
}
#endif

void XMLDoc::dump( std::ostream& os )
{
	os << "@ " << typeid( this ).name() << std::endl << (const char*)stream() << std::endl;
}

void XMLDoc::dump( Socket& skt )
{
	skt << "@ " << typeid( this ).name() << "\n" << (const char*)stream() << "\n";
}
